package com.tzld.piaoquan.content.understanding.rocketmq.consumer;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.TypeReference;
import com.tzld.piaoquan.content.understanding.common.enums.ExceptionEnum;
import com.tzld.piaoquan.content.understanding.common.enums.TaskStatusEnum;
import com.tzld.piaoquan.content.understanding.common.exception.CommonException;
import com.tzld.piaoquan.content.understanding.dao.mapper.CuPipelineMapper;
import com.tzld.piaoquan.content.understanding.dao.mapper.CuPipelineStepMapper;
import com.tzld.piaoquan.content.understanding.dao.mapper.CuTaskMapper;
import com.tzld.piaoquan.content.understanding.model.dto.ContentAnalyseDTO;
import com.tzld.piaoquan.content.understanding.model.dto.PipelineTaskMessageDTO;
import com.tzld.piaoquan.content.understanding.model.dto.TreeNode;
import com.tzld.piaoquan.content.understanding.model.po.*;
import com.tzld.piaoquan.content.understanding.service.ContentService;
import com.tzld.piaoquan.content.understanding.service.PipelineService;
import com.tzld.piaoquan.content.understanding.util.RedisUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.rocketmq.client.annotation.RocketMQMessageListener;
import org.apache.rocketmq.client.apis.consumer.ConsumeResult;
import org.apache.rocketmq.client.apis.message.MessageId;
import org.apache.rocketmq.client.apis.message.MessageView;
import org.apache.rocketmq.client.core.RocketMQListener;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

/**
 * 旧版视频理解 需要延迟处理的视频 消费者
 *
 * @author supeng
 */
@Slf4j
@Component
@RocketMQMessageListener(endpoints = "${rocketmq.consumer.endpoints:rmq-cn-vym47zv2i03-vpc.cn-hangzhou.rmq.aliyuncs.com:8080}",
        topic = "${rocketmq.consumer.pipelinetask.topic:topic_content_understanding_pipeline_task}",
        consumerGroup = "${rocketmq.consumer.pipelinetask.group:group_content_understanding_pipeline_task}", tag = "*", consumptionThreadCount = 8)
public class ContentUnderstandingPipelineTaskConsumer implements RocketMQListener {

    @Autowired
    private PipelineService pipelineService;

    @Autowired
    private CuTaskMapper cuTaskMapper;

    @Autowired
    private CuPipelineMapper cuPipelineMapper;
    @Autowired
    private CuPipelineStepMapper cuPipelineStepMapper;
    @Value("${same.video.understanding.skip.seconds:0L}")
    private Long sameVideoUnderstandingSkipSeconds;
    @Autowired
    private RedisUtil redisUtil;
    /**
     * 相同视频跳过理解
     * same:video:understanding:skip:{videoId}
     */
    private static final String SAME_VIDEO_SKIP_KEY = "same:video:understanding:skip:%s";

    @Override
    public ConsumeResult consume(MessageView messageView) {
        log.info("Receive message {}", messageView);
        MessageId messageId = messageView.getMessageId();
        try {
            // 1.解析消息
            ByteBuffer body = messageView.getBody();
            Charset charset = StandardCharsets.UTF_8;
            String messageStr = charset.decode(body).toString();
            PipelineTaskMessageDTO messageDTO = JSON.parseObject(messageStr, PipelineTaskMessageDTO.class);
            String taskId = messageDTO.getTaskId();
            Long videoId = messageDTO.getVideoId();
            Map<String, Object> extMap = messageDTO.getExtMap();
            if (extMap == null) {
                extMap = new HashMap<>();
            }
            extMap.put("messageId", messageId.toString());
            CuTaskExample example = new CuTaskExample();
            example.createCriteria().andTaskIdEqualTo(taskId).andTaskStatusEqualTo(TaskStatusEnum.SUBMIT.getValue());
            List<CuTask> list = cuTaskMapper.selectByExample(example);
            if (Objects.isNull(list) || list.isEmpty()) {
                //无需处理
                log.info("数据无效 taskId = {} videoId = {}", taskId, videoId);
                return ConsumeResult.SUCCESS;
            }
            //相同视频跳过
            if (Objects.nonNull(sameVideoUnderstandingSkipSeconds) && sameVideoUnderstandingSkipSeconds > 0) {
                String key = String.format(SAME_VIDEO_SKIP_KEY, videoId);
                Boolean flag = redisUtil.setNx(key, "1", sameVideoUnderstandingSkipSeconds);
                if (Objects.isNull(flag) || !flag) {
                    log.info("相同视频跳过理解 videoId = {} taskId = {} sameVideoUnderstandingSkipSeconds = {}", videoId, taskId, sameVideoUnderstandingSkipSeconds);
                    String reason = "相同视频跳过理解，时间间隔：" + sameVideoUnderstandingSkipSeconds;
                    pipelineService.updateTaskResult(taskId, null, reason, videoId, extMap, false);
                    return ConsumeResult.SUCCESS;
                }
            }
            CuTask cuTask = list.get(0);
            Long pipelineId = cuTask.getPipelineId();
            String input = cuTask.getInput();
            if (input == null || input.isEmpty()) {
                log.info("输入内容为空 taskId = {} videoId = {} pipelineId = {}", taskId, videoId, pipelineId);
                String reason = "输入内容为空";
                pipelineService.updateTaskResult(taskId, null, reason, videoId, extMap, false);
                return ConsumeResult.SUCCESS;
            }
            CuPipeline cuPipeline = cuPipelineMapper.selectByPrimaryKey(pipelineId);
            if (Objects.isNull(cuPipeline)) {
                //pipeline 无数据
                log.info("输入内容为空 taskId = {} videoId = {} pipelineId = {}", taskId, videoId, pipelineId);
                String reason = "pipelineId[" + pipelineId + "]无效";
                pipelineService.updateTaskResult(taskId, null, reason, videoId, extMap, false);
                return ConsumeResult.SUCCESS;
            }
            // 2.获取pipeline配置
            Integer contentType = cuPipeline.getContentType();
            CuPipelineStepExample stepExample = new CuPipelineStepExample();
            stepExample.createCriteria().andPipelineIdEqualTo(pipelineId);
            List<CuPipelineStep> stepList = cuPipelineStepMapper.selectByExample(stepExample);
            if (Objects.isNull(stepList) || stepList.isEmpty()) {
                log.info("输入内容为空 taskId = {} videoId = {} pipelineId = {}", taskId, videoId, pipelineId);
                String reason = "pipelineId[" + pipelineId + "]无步骤配置";
                pipelineService.updateTaskResult(taskId, null, reason, videoId, extMap, false);
                return ConsumeResult.SUCCESS;
            }
            // 3.构建tree结构
            TreeNode root = pipelineService.buildTree(stepList);
            root.setType(contentType);
            root.setInput(input);
            ContentAnalyseDTO dto1 = new ContentAnalyseDTO();
            dto1.setTaskId(taskId);
            dto1.setVideoId(videoId);
            dto1.setContent(input);
            dto1.setExtMap(extMap);
            // 4.按照步骤执行pipeline每一步动作
            pipelineService.executeTaskTreeNodeBFS(root, dto1);
            return ConsumeResult.SUCCESS;
        } catch (Exception e) {
            log.error("Message is failed to be acknowledged, messageId={}", messageId, e);
        }
        return ConsumeResult.FAILURE;
    }
}
