package com.tzld.piaoquan.content.understanding.service.impl;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.TypeReference;
import com.tzld.piaoquan.content.understanding.common.base.CommonResponse;
import com.tzld.piaoquan.content.understanding.common.enums.AigcServerExceptionEnum;
import com.tzld.piaoquan.content.understanding.common.enums.ContentTypeEnum;
import com.tzld.piaoquan.content.understanding.common.exception.StopRetryException;
import com.tzld.piaoquan.content.understanding.model.param.ActionParam;
import com.tzld.piaoquan.content.understanding.model.param.GeminiParam;
import com.tzld.piaoquan.content.understanding.service.Action;
import com.tzld.piaoquan.content.understanding.util.HttpClientUtil;
import com.tzld.piaoquan.content.understanding.util.HttpPoolClient;
import com.tzld.piaoquan.content.understanding.util.RedisUtil;
import com.tzld.piaoquan.content.understanding.util.RetryUtil;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

/**
 * @author supeng
 */
@Slf4j
@Service(value = "geminiGenerateContentAction")
public class GeminiGenerateContentAction implements Action {

    @Value("${ai.geminiApi.url:http://ai-api.piaoquantv.com/aigc-server/gemini/generateContent}")
    private String geminiApiUrl;

    @Value("#{'${gemini.apikey.list:[]}'.split(',')}")
    private List<String> apiKeyList;

    private static final HttpPoolClient httpPoolClient = HttpClientUtil.create(3000, 1200000, 50, 200, 0, 3000);

    private static final String MODEL = "gemini-2.0-flash";
    private static final Double TEMPERATURE = 1.0;

    private static final Integer RETRY_TIMES = 3;

    @Autowired
    private RedisUtil redisUtil;

    private static final String GEMINI_API_KEY_INDEX = "gemini:apikey:list.index";

    @Override
    public String execute(ActionParam param) {
        Integer type = param.getType();
        String apiKey = "";
        int size;
        boolean useLocalApiKey;
        if (Objects.nonNull(apiKeyList) && !apiKeyList.isEmpty()) {
            useLocalApiKey = true;
            size = apiKeyList.size();
            int index = getIndex(size);
            apiKey = apiKeyList.get(index);
        } else {
            useLocalApiKey = false;
            size = 0;
        }
        GeminiParam geminiParam = new GeminiParam();
        geminiParam.setType(type);
        geminiParam.setPrompt(param.getPrompt());
        if (Objects.equals(ContentTypeEnum.COVER.getValue(), type) || Objects.equals(ContentTypeEnum.VIDEO.getValue(), type)
                || Objects.equals(ContentTypeEnum.AUDIO.getValue(), type) || Objects.equals(ContentTypeEnum.SRT.getValue(), type)
                || Objects.equals(ContentTypeEnum.VTT.getValue(), type)) {
            geminiParam.setMediaUrl(param.getInput());
        }
        geminiParam.setApiKey(apiKey);
        geminiParam.setModel(MODEL);
        geminiParam.setTemperature(TEMPERATURE);
        return RetryUtil.executeWithRetry(() -> {
            try {
                log.info("geminiGenerateContentAction geminiParam = {}", JSON.toJSONString(geminiParam));
                Optional<String> optionalS = httpPoolClient.postJson(geminiApiUrl, JSON.toJSONString(geminiParam));
                log.info("geminiGenerateContentAction optionalS = {}", optionalS);
                if (optionalS.isPresent()) {
                    CommonResponse<Map<String, Object>> commonResponse = JSON.parseObject(optionalS.get(), new TypeReference<CommonResponse<Map<String, Object>>>() {
                    });
                    if (commonResponse.isSuccess() && Objects.nonNull(commonResponse.getData()) && Objects.nonNull(commonResponse.getData().get("result"))) {
                        String content = commonResponse.getData().get("result").toString();
                        //过滤一些特殊格式
                        content = content.replace("```json", "").replace("```", "");
                        return content;
                    } else if (AigcServerExceptionEnum.PARAM_ERROR.getCode() == commonResponse.getCode()) {
                        throw new StopRetryException("geminiGenerateContentAction request params error geminiParam = " + JSON.toJSONString(geminiParam) + ", response = " + JSON.toJSONString(commonResponse));
                    } else {
                        if (useLocalApiKey) {
                            int index1 = getIndex(size);
                            String apiKey1 = apiKeyList.get(index1);
                            geminiParam.setApiKey(apiKey1);
                        }
                        throw new RuntimeException("geminiGenerateContentAction request failure geminiParam = " + JSON.toJSONString(geminiParam));
                    }
                } else {
                    if (useLocalApiKey) {
                        int index1 = getIndex(size);
                        String apiKey1 = apiKeyList.get(index1);
                        geminiParam.setApiKey(apiKey1);
                    }
                    throw new RuntimeException("geminiGenerateContentAction optionalS is null geminiParam = " + JSON.toJSONString(geminiParam));
                }
            } catch (Exception e) {
                if (useLocalApiKey) {
                    int index1 = getIndex(size);
                    String apiKey1 = apiKeyList.get(index1);
                    geminiParam.setApiKey(apiKey1);
                    log.error("geminiGenerateContentAction error param = {}", JSON.toJSONString(geminiParam), e);
                }
                throw new RuntimeException("geminiGenerateContentAction request error", e);
            }
        }, RETRY_TIMES, "geminiGenerateContentAction", RetryUtil::retryInterval);
    }

    private int getIndex(int size) {
        int index = 0;
        Long value = redisUtil.incr(GEMINI_API_KEY_INDEX);
        if (Objects.nonNull(value)) {
            long indexValue = value - 1;
            index = (int) (indexValue % size);
            if (value > 0 && value % size == 0) {
                try {
                    redisUtil.deleteKey(GEMINI_API_KEY_INDEX);
                } catch (Exception e) {
                    // 重置失败不影响当前获取元素的操作，只记录错误
                    log.error(e.getMessage(), e);
                }
            }
        }
        return index;
    }

}
