package com.tzld.piaoquan.content.understanding.service.impl;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.TypeReference;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import com.tzld.piaoquan.content.understanding.common.enums.ExceptionEnum;
import com.tzld.piaoquan.content.understanding.common.enums.TaskStatusEnum;
import com.tzld.piaoquan.content.understanding.common.exception.CommonException;
import com.tzld.piaoquan.content.understanding.dao.mapper.CuPipelineMapper;
import com.tzld.piaoquan.content.understanding.dao.mapper.CuPipelineStepMapper;
import com.tzld.piaoquan.content.understanding.dao.mapper.CuPromptMapper;
import com.tzld.piaoquan.content.understanding.dao.mapper.CuTaskMapper;
import com.tzld.piaoquan.content.understanding.model.dto.ContentAnalyseDTO;
import com.tzld.piaoquan.content.understanding.model.dto.ContentDTO;
import com.tzld.piaoquan.content.understanding.model.dto.PipelineTaskMessageDTO;
import com.tzld.piaoquan.content.understanding.model.dto.TreeNode;
import com.tzld.piaoquan.content.understanding.model.param.ActionParam;
import com.tzld.piaoquan.content.understanding.model.param.ContentAnalyseParam;
import com.tzld.piaoquan.content.understanding.model.po.*;
import com.tzld.piaoquan.content.understanding.rocketmq.producer.ContentUnderstandingPipelineTaskProducer;
import com.tzld.piaoquan.content.understanding.service.Action;
import com.tzld.piaoquan.content.understanding.service.LoghubService;
import com.tzld.piaoquan.content.understanding.service.PipelineService;
import lombok.extern.slf4j.Slf4j;
import org.apache.rocketmq.client.apis.message.MessageId;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import java.util.*;
import java.util.concurrent.*;

/**
 * @author supeng
 */
@Slf4j
@Service
public class PipelineServiceImpl implements PipelineService {

    @Autowired
    private CuPipelineMapper cuPipelineMapper;

    @Autowired
    private CuPipelineStepMapper cuPipelineStepMapper;

    @Autowired
    private CuPromptMapper cuPromptMapper;

    @Autowired
    private Map<String, Action> actionMap;

    @Autowired
    private MongoTemplate mongoTemplate;

    private static final String COLLECTION_NAME = "content_understanding_info";

    @Autowired
    private CuTaskMapper cuTaskMapper;

    @Autowired
    private LoghubService loghubService;

    @Value("${aliyun.log.project}")
    private String aliyunLogProject;
    @Value("${aliyun.log.logstore.result}")
    private String aliyunLogStoreResult;

    @Autowired
    private ContentUnderstandingPipelineTaskProducer contentUnderstandingPipelineTaskProducer;

    /**
     * 线程池队列大小
     */
    private static final int QUEUE_MAX_SIZE = 100000;
    /**
     * 线程命名
     */
    private static final ThreadFactory NAMED_THREAD_FACTORY = new ThreadFactoryBuilder().setNameFormat("pipeline-service-pool-%d").build();
    /**
     * 线程池
     */
    private static ExecutorService pool;

    @PostConstruct
    public void init() {
        //init thread pool
        pool = new ThreadPoolExecutor(32, 32,
                0L, TimeUnit.MILLISECONDS, new LinkedBlockingQueue<>(QUEUE_MAX_SIZE), NAMED_THREAD_FACTORY, new ThreadPoolExecutor.AbortPolicy());
    }

    @PreDestroy
    public void destroy() {
        //gracefully shutdown
        pool.shutdown();
    }

    @Override
    public void execute(ContentAnalyseParam param) {
        // 1.获取pipeline配置
        Long pipelineId = param.getPipelineId();
        String content = param.getContent();
        Integer type = param.getType();
        CuPipelineStepExample example = new CuPipelineStepExample();
        example.createCriteria().andPipelineIdEqualTo(pipelineId);
        List<CuPipelineStep> stepList = cuPipelineStepMapper.selectByExample(example);
        if (Objects.isNull(stepList) || stepList.isEmpty()) {
            throw new CommonException(ExceptionEnum.CONFIG_ERROR, "无配置 pipelineId：" + pipelineId);
        }
        // 2.构建tree结构
        TreeNode root = buildTree(stepList);
        root.setType(type);
        root.setInput(content);
        ContentAnalyseDTO dto = new ContentAnalyseDTO();
        BeanUtils.copyProperties(param, dto);
        // 3.按照步骤执行pipeline每一步动作
        executeTreeNodeBFS(root, dto);
    }

    @Override
    public String executeTask(ContentDTO contentDTO) {
        // 1.数据准备
        Long pipelineId = contentDTO.getPipelineId();
        String content = contentDTO.getContent();
        // 1.新增任务
        String taskId = UUID.randomUUID().toString().replace("-", "") + System.currentTimeMillis();
        CuTask cuTask = new CuTask();
        cuTask.setTaskId(taskId);
        cuTask.setPipelineId(pipelineId);
        cuTask.setInput(content);
        int insertCount = cuTaskMapper.insertSelective(cuTask);
        if (insertCount <= 0) {
            throw new CommonException(ExceptionEnum.SYSTEM_ERROR, "新增任务失败");
        }
        // 2.获取pipeline配置
        Integer contentType = contentDTO.getContentType();
        CuPipelineStepExample example = new CuPipelineStepExample();
        example.createCriteria().andPipelineIdEqualTo(pipelineId);
        List<CuPipelineStep> stepList = cuPipelineStepMapper.selectByExample(example);
        if (Objects.isNull(stepList) || stepList.isEmpty()) {
            throw new CommonException(ExceptionEnum.CONFIG_ERROR, "无配置 pipelineId：" + pipelineId);
        }
        if (Objects.isNull(content)) {
            throw new CommonException(ExceptionEnum.DATA_ERROR, "数据异常content：" + content);
        }
        //异步执行
        pool.execute(() -> {
            // 3.构建tree结构
            TreeNode root = buildTree(stepList);
            root.setType(contentType);
            root.setInput(content);
            ContentAnalyseDTO dto1 = new ContentAnalyseDTO();
            dto1.setTaskId(taskId);
            dto1.setVideoId(contentDTO.getVideoId());
            dto1.setContent(content);
            dto1.setExtMap(contentDTO.getExtMap());
            // 4.按照步骤执行pipeline每一步动作
            executeTaskTreeNodeBFS(root, dto1);
        });
        return taskId;
    }

    @Override
    public String submitTask(ContentDTO contentDTO) {
        // 1.数据准备
        Long pipelineId = contentDTO.getPipelineId();
        Long videoId = contentDTO.getVideoId();
        String content = contentDTO.getContent();
        // 1.新增任务
        String taskId = UUID.randomUUID().toString().replace("-", "") + System.currentTimeMillis();
        CuTask cuTask = new CuTask();
        cuTask.setTaskId(taskId);
        cuTask.setPipelineId(pipelineId);
        cuTask.setInput(content);
        int insertCount = cuTaskMapper.insertSelective(cuTask);
        if (insertCount <= 0) {
            throw new CommonException(ExceptionEnum.SYSTEM_ERROR, "新增任务失败");
        }
        PipelineTaskMessageDTO messageDTO = new PipelineTaskMessageDTO();
        messageDTO.setTaskId(taskId);
        messageDTO.setVideoId(videoId);
        messageDTO.setExtMap(contentDTO.getExtMap());
        try {
            MessageId messageId = contentUnderstandingPipelineTaskProducer.syncSendNormalMessage(messageDTO);
            log.info("syncSendNormalMessage messageId = {}", messageId);
        } catch (Exception e) {
            log.error("syncSendNormalMessage error messageDTO = {}", e, messageDTO);
        }
        return taskId;
    }

    @Override
    public TreeNode buildTree(List<CuPipelineStep> stepList) {
        Map<Long, TreeNode> nodeMap = new HashMap<>();
        TreeNode root = null;

        // 将所有节点放入 Map 中，方便后续查找
        for (CuPipelineStep step : stepList) {
            TreeNode node = new TreeNode();
            node.setStep(step);
            nodeMap.put(step.getId(), node);
        }

        // 构建父子关系
        for (CuPipelineStep step : stepList) {
            Long parentId = step.getParentId();
            if (Objects.isNull(parentId) || parentId == 0) {
                root = nodeMap.get(step.getId());
            } else if (nodeMap.containsKey(parentId)) {
                TreeNode parentNode = nodeMap.get(parentId);
                TreeNode currentNode = nodeMap.get(step.getId());
                List<TreeNode> children = parentNode.getChildren();
                if (Objects.isNull(children)) {
                    children = new ArrayList<>();
                    parentNode.setChildren(children);
                }
                children.add(currentNode);
            }
        }
        return root;
    }

    /**
     * 遍历执行
     *
     * @param root 根节点
     * @param dto  数据
     */
    public void executeTreeNodeBFS(TreeNode root, ContentAnalyseDTO dto) {
        if (Objects.isNull(root) || Objects.isNull(dto) || Objects.isNull(dto.getVideoId())) {
            return;
        }
        Queue<TreeNode> queue = new LinkedList<>();
        queue.offer(root);

        Long videoId = dto.getVideoId();

        while (!queue.isEmpty()) {
            TreeNode currentNode = queue.poll();
            CuPipelineStep step = currentNode.getStep();
            Integer type = currentNode.getType();
            String input = currentNode.getInput() == null ? "" : currentNode.getInput();
            // 执行action
            String result = executeStep(step, input, type);
            if (Objects.isNull(currentNode.getChildren()) || currentNode.getChildren().isEmpty()) {
                if (Objects.nonNull(result)) {
                    //异常不阻塞其他分支执行
                    try {
                        Map<String, Object> resultMap = JSON.parseObject(result, new TypeReference<Map<String, Object>>() {
                        });
                        resultMap.put("video_id", videoId);
                        mongoTemplate.insert(resultMap, COLLECTION_NAME);
                    } catch (Exception e) {
                        log.error("step execute error step = {}, result = {}", e, JSON.toJSONString(step), result);
                    }
                }
                continue;
            }
            for (TreeNode child : currentNode.getChildren()) {
                child.setInput(result);
                queue.offer(child);
            }
        }
    }

    /**
     * 遍历执行
     *
     * @param root 根节点
     * @param dto  数据
     */
    @Override
    public void executeTaskTreeNodeBFS(TreeNode root, ContentAnalyseDTO dto) {
        if (Objects.isNull(root) || Objects.isNull(dto) || Objects.isNull(dto.getVideoId())) {
            return;
        }
        Queue<TreeNode> queue = new LinkedList<>();
        queue.offer(root);

        String taskId = dto.getTaskId();
        Long videoId = dto.getVideoId();

        while (!queue.isEmpty()) {
            TreeNode currentNode = queue.poll();
            CuPipelineStep step = currentNode.getStep();
            Integer type = currentNode.getType();
            String input = currentNode.getInput() == null ? "" : currentNode.getInput();
            String reason = "处理失败";
//            String result = null;
//            try {
//                // 执行action
//                result = executeTaskStep(step, input, type);
//            } catch (Exception e) {
//                reason = e.getMessage();
//                log.error("step execute error step = {}, reason = {}", e, JSON.toJSONString(step), reason);
//            }
            String result = executeTaskStep(step, input, type);
            if (Objects.isNull(currentNode.getChildren()) || currentNode.getChildren().isEmpty()) {
                if (Objects.nonNull(result)) {
                    //异常不阻塞其他分支执行
                    try {
                        //替换空格
                        result = result.replace(" ", "");
                        //保存数据到mysql
                        updateTaskResult(taskId, result, reason, videoId, dto.getExtMap(), true);
                    } catch (Exception e) {
                        log.error("step execute update error step = {}, result = {}", e, JSON.toJSONString(step), result);
                    }
                } else {
                    updateTaskResult(taskId, result, reason, videoId, dto.getExtMap(), false);
                }
                continue;
            }
            for (TreeNode child : currentNode.getChildren()) {
                child.setInput(result);
                queue.offer(child);
            }
        }
    }

    @Override
    public void updateTaskResult(String taskId, String result, String reason, Long videoId, Map<String, Object> extMap, boolean isSuccess) {
        CuTaskExample example = new CuTaskExample();
        example.createCriteria().andTaskIdEqualTo(taskId);
        CuTask cuTask = new CuTask();
        if (isSuccess) {
            cuTask.setOutput(result);
            cuTask.setTaskStatus(TaskStatusEnum.SUCCESS.getValue());
        } else {
            cuTask.setReason(reason);
            cuTask.setTaskStatus(TaskStatusEnum.FAILURE.getValue());
        }
        int updateCount = cuTaskMapper.updateByExampleSelective(cuTask, example);
        if (updateCount <= 0) {
            log.error("updateTaskResult execute update error taskId = {} videoId = {}, result = {} reason = {} updateCount = {}", taskId, videoId, result, reason, updateCount);
        }
        Map<String, Object> logMap = new HashMap<>();
        logMap.put("video_id", videoId);
        logMap.put("taskId", taskId);
        if (isSuccess) {
            logMap.put("data", result);
            logMap.put("status", "success");
        } else {
            logMap.put("error_msg", reason);
            logMap.put("status", "failure");
        }
        if (Objects.nonNull(extMap)) {
            logMap.putAll(extMap);
        }
        loghubService.asyncSubmitLog(aliyunLogProject, aliyunLogStoreResult, "", logMap);
    }

    /**
     * 执行step 动作
     *
     * @param step
     * @param input
     * @return
     */
    public String executeStep(CuPipelineStep step, String input, Integer type) {
        CuPrompt cuPrompt = cuPromptMapper.selectByPrimaryKey(step.getPromptId());
        if (Objects.isNull(cuPrompt) || Objects.isNull(cuPrompt.getId())) {
            return null;
        }
        String prompt = String.format(cuPrompt.getPrompt(), input);

        Action action = actionMap.get(step.getAction());
        ActionParam actionParam = new ActionParam();
        actionParam.setType(type);
        actionParam.setInput(input);
        actionParam.setPrompt(prompt);
        return action.execute(actionParam);
    }

    /**
     * 执行step 动作
     *
     * @param step
     * @param input
     * @return
     */
    public String executeTaskStep(CuPipelineStep step, String input, Integer type) {
        CuPrompt cuPrompt = cuPromptMapper.selectByPrimaryKey(step.getPromptId());
        if (Objects.isNull(cuPrompt) || Objects.isNull(cuPrompt.getId())) {
            return null;
        }
        String prompt = String.format(cuPrompt.getPrompt(), input);

        Action action = actionMap.get(step.getAction());
        ActionParam actionParam = new ActionParam();
        actionParam.setType(type);
        actionParam.setInput(input);
        actionParam.setPrompt(prompt);
        return action.execute(actionParam);
    }
}
