package com.tzld.piaoquan.content.understanding.util;

import com.tzld.piaoquan.content.understanding.common.exception.StopRetryException;
import lombok.extern.slf4j.Slf4j;

import java.util.Objects;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;
import java.util.function.LongSupplier;
import java.util.function.Predicate;
import java.util.function.Supplier;

@Slf4j
public class RetryUtil {

    private static final Integer MAX_RETRIES = 3;

    /**
     * 重试
     *
     * @param action
     * @param actionName
     * @param <T>
     * @return
     */
    public static <T> T executeWithRetry(Supplier<T> action, String actionName) {
        return executeWithRetry(action, MAX_RETRIES, actionName, null);
    }

    /**
     * 重试
     *
     * @param action
     * @param maxRetries
     * @param actionName
     * @param <T>
     * @return
     */
    public static <T> T executeWithRetry(Supplier<T> action, Integer maxRetries, String actionName) {
        return executeWithRetry(action, maxRetries, actionName, null);
    }

    /**
     * 重试
     *
     * @param action
     * @param maxRetries
     * @param actionName
     * @param retryIntervalProvider 毫秒
     * @param <T>
     * @return
     */
    public static <T> T executeWithRetry(Supplier<T> action, Integer maxRetries, String actionName, Function<Integer, Long> retryIntervalProvider) {
        for (int i = 1; i <= maxRetries; i++) {
            try {
                return action.get();
            } catch (Exception e) {
                log.error("{} failed on attempt {}: {}", actionName, i, e.getMessage(), e);
                if (e instanceof StopRetryException) {
                    log.error("{} failed on attempt {} due to StopRetryException. Stopping retries.", actionName, i);
                    throw new RuntimeException(actionName + " failed due to Stopping retries.", e);
                }
                if (i < maxRetries) {
                    if (Objects.nonNull(retryIntervalProvider) && Objects.nonNull(retryIntervalProvider.apply(i))) {
                        long interval = retryIntervalProvider.apply(i);
                        if (interval > 0) {
                            try {
                                TimeUnit.MILLISECONDS.sleep(interval);
                            } catch (InterruptedException ex) {
                                log.error("{} failed on attempt {}: {}", actionName, i, ex.getMessage(), ex);
                            }
                        }
                    }
                    log.info("{} will be retried (attempt {}/{})", actionName, i + 1, maxRetries);
                } else {
                    log.error("{} failed after {} retries due to non-IO error.", actionName, maxRetries);
                    throw new RuntimeException(actionName + " failed after multiple retries due to error.", e);
                }
            }
        }
        return null;
    }

    /**
     * @param i 第i次
     * @return 毫秒
     */
    public static long retryInterval(int i) {
        if (i == 1) {
            return 5 * 1000L;
        }
        if (i == 2) {
            return 10 * 1000L;
        }
        if (i == 3) {
            return 30 * 1000L;
        }
        if (i == 4) {
            return 60 * 1000L;
        }
        if (i == 5) {
            return 2 * 60 * 1000L;
        }
        if (i == 6) {
            return 3 * 60 * 1000L;
        }
        if (i == 7) {
            return 5 * 60 * 1000L;
        }
        if (i == 8) {
            return 10 * 60 * 1000L;
        }
        if (i == 9) {
            return 20 * 60 * 1000L;
        }
        if (i == 10) {
            return 30 * 60 * 1000L;
        }

        return 0L;
    }

    public static long retryIntervalLevel1(int i) {
        if (i == 1) {
            return 1000L;
        }
        if (i == 2) {
            return 3 * 1000L;
        }
        if (i == 3) {
            return 5 * 1000L;
        }
        if (i == 4) {
            return 8 * 1000L;
        }
        if (i == 5) {
            return 10 * 1000L;
        }
        if (i == 6) {
            return 10 * 1000L;
        }
        if (i == 8) {
            return 20 * 1000L;
        }
        if (i == 9) {
            return 20 * 1000L;
        }
        if (i == 10) {
            return 20 * 1000L;
        }
        return 0L;
    }
}
